// Copyright Shareef Jalloq
// SPDX-License-Identifier: MIT
//
// Auto-generated by peakrdl-sv

<%
  lname = block.inst_name.lower()
  ublock = block.inst_name.upper()

  addr_width = block.addrwidth
  data_width = block.accesswidth
  registers  = block.get_registers()
  num_regs   = len(registers)
  max_regs_char = len("{}".format(num_regs-1))

%>\
package ${lname}_reg_pkg;

  // Address width
  parameter int BlockAw = ${block.addrwidth};


  % for r in registers:
  // ${r.path}
  typedef struct packed {
  % if len(r) == 1:
    logic ${sv_bitarray(r[0])} q;
    % if r[0].swmod:
    logic qe;
    % endif
  % else:
    % for f in r:
    struct packed {
      logic ${sv_bitarray(f)} q;
      % if f.swmod:
      logic qe;
      % endif
    } ${f.name};
    % endfor
  % endif
  } ${r.owning_addrmap.inst_name}_reg2hw_${r.path}_t;

  % endfor\

  % for r in registers:
  // ${r.path}
  typedef struct packed {
    % if len(r) == 1:
    logic ${sv_bitarray(r[0])} d;
    logic de;
    % else:
    % for f in r:
    struct packed {
      logic ${sv_bitarray(f)} d;
      logic de;
    } ${f.name};
    % endfor
    % endif
  } ${hw2reg_t_gen(r)};

  % endfor

  // Register -> HW
  typedef struct packed {
  % for r in registers:
    ${reg2hw_t_gen(r)} ${r.path};
  % endfor 
  } ${lname}_reg2hw_t;

  // HW -> Register
  typedef struct packed {
  % for r in registers:
    ${hw2reg_t_gen(r)} ${r.path};
  % endfor 
  } ${lname}_hw2reg_t;

  // Register address offsets
  % for r in registers:
<%
    if r.is_wide:
      enables = r.regwidth // r.accesswidth
      address_base = r.absolute_address
      address_incr = r.accesswidth // 8
      values = []
      params = []
      for i in range(enables):
        values.append(f"{addr_width}'h{address_base+(i*address_incr):X}")
        params.append(f"{ublock}_{r.path.upper()}_{i}_OFFSET")
    else:
      values = [f"{addr_width}'h{r.absolute_address:X}"]
      params = [f"{ublock}_{r.path.upper()}_OFFSET"]
%>\
  % for p,v in zip(params, values):
  parameter logic [BlockAw-1:0] ${p} = ${v};
  % endfor
  % endfor

endpackage
<%def name="hw2reg_t_gen(reg)" filter="trim">\
${reg.owning_addrmap.inst_name}_hw2reg_${reg.path}_t
</%def>\
<%def name="reg2hw_t_gen(reg)" filter="trim">\
${reg.owning_addrmap.inst_name}_reg2hw_${reg.path}_t
</%def>\
<%def name="sv_bitarray(field)" filter="trim">\
% if field.width > 1:
[${field.width-1}:0]
% endif
</%def>
